/* an example of calling the NAG pseudorandom generator nag_rand_basic (g05sac)
   in an OpenMP multi-threaded environment */
#include <stdio.h>
#include <math.h>
#include <nag.h>
#include <nagx04.h>
#include <nag_stdlib.h>
#include <nagg05.h>
#include <omp.h>

int main(int argc, char *argv[]) {
  int i, exit_status = 0;
  Integer NumberOfThreads, iskip, tskip, ThreadNo, npaths, tnpaths, t1, lstate;
  double sumRandom;
  Integer *state = 0;
  Integer seed[1];
  double *x = 0;  
  NagError fail;

  /* use the Wichmann Hill II geneator:
     at Mark 9 of the C library skip ahead is only available for 4 of the 
     available 5 generators, the exception being the Mersenne Twister. Skip 
     ahead for the Mersenne Twister will appear in the next Mark */
  Nag_BaseRNG genid = Nag_WichmannHill_II;
  Integer subid = 0;

  /* only using one seed value */
  Integer lseed = 1;

  /* parse the input arguments */
  if (argc < 3) {
    printf("Usage: XXX n i s\n");
    printf("Where:\n");
    printf("  n = number of paths\n");
    printf("  i = initial skip\n");
    printf("  s = seed (optional)\n");
    return 0;
  }

  /* get the problem size */
  npaths = atoi(argv[1]);
  iskip = atoi(argv[2]);
  if (argc > 3) {
    /* use the supplied value for the seed */
    seed[0] = atoi(argv[3]);
  } else {
    /* use a default value for the seed */
    seed[0] = 1762543;
  }

  /* intialise the error structure */
  INIT_FAIL(fail);

  /* get the length of the state array for the psuedo-generator */
  lstate = -1;
  nag_rand_init_repeatable(genid, subid, seed, lseed, state, &lstate, &fail);
  if (fail.code != NE_NOERROR)
    {
      printf("Error from nag_rand_init_repeatable (g05kfc).\n%s\n", 
             fail.message);
      exit_status = 1;
      goto END;
    }

#pragma omp parallel default(none) \
  reduction(+:sumRandom) \
    private(ThreadNo,tskip,tnpaths,x,fail,i,t1,NumberOfThreads,state) \
      shared(npaths,exit_status,iskip,genid,subid,lstate,seed,lseed)
  {

    /* thread number */
    ThreadNo = omp_get_thread_num();

    /* number of threads */
    NumberOfThreads = omp_get_num_threads();

    /* calculate the skip and the number of values to generate for this thread */
    if (NumberOfThreads == 1) {
      tskip = 0;
      tnpaths = npaths;
    } else {
      t1 = npaths % NumberOfThreads;
      t1 = (ThreadNo < t1) ? ThreadNo : t1;
      tskip = ThreadNo * (npaths / NumberOfThreads) + t1;
      tnpaths = (npaths + NumberOfThreads - ThreadNo - 1) / NumberOfThreads;
    } 

    /* adjust the skip for this thread by the initial skip */
    tskip += iskip;

    /* intialise the error structure */
    INIT_FAIL(fail);

    /* allocate memory to hold the generated values (x) and state of the pseudo 
       generator (state) */
    if (!(x = NAG_ALLOC(tnpaths, double)) ||
        !(state = NAG_ALLOC(lstate,Integer)))
      {
        printf("Allocation failure in thread %"NAG_IFMT"\n", ThreadNo + 1);
        exit_status = -1;
        goto TEND;
      }
    
    /* nag_rand_init_repeatable (g05kfc): initialise the pseudo generator, 
       all threads need to be initialised using the same seed */
    nag_rand_init_repeatable(genid, subid, seed, lseed, state, &lstate, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_rand_init_repeatable (g05kfc) in thread "
               "%"NAG_IFMT".\n%s\n", ThreadNo + 1, fail.message);
        exit_status = 1;
        goto TEND;
      }

    /* nag_rand_skip_ahead (g05kjc): skip ahead the sequence tskip places on 
       each thread */
    nag_rand_skip_ahead(tskip, state, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_rand_skip_ahead (g05kjc).\n%s\n", fail.message);
        exit_status = 1;
        goto TEND;
      }
    
    /* nag_rand_basic (g05sac): generate tnpaths values from a uniform 
       distribution */
    nag_rand_basic(tnpaths, state, x, &fail);
    if (fail.code != NE_NOERROR)
      {
        printf("Error from nag_rand_basic (g05sac).\n%s\n", fail.message);
        exit_status = 1;
        goto TEND;
      }
    
    /* calculate the sum for this particular thread */    
    for (i = 0; i < tnpaths; i++)
      sumRandom += x[i];

    printf("Thread = %4"NAG_IFMT", Sum = %g\n", ThreadNo+1,sumRandom);
    
    /* free memory */
  TEND:
    if (x) free(x);
    if (state) free(state);
  }
  
  /* display the results */
  if (!exit_status) {
    printf("\nTotal Sum = %g\n", sumRandom);
  }

 END:
  return exit_status;
}
