    PROGRAM fpseudo_openmp
!      NAG Copyright 2011.
!      an example of calling the NAG pseudorandom generator G05SAF in an OpenMP
!      multi-threaded environment

!      .. Use Statements ..
       USE nag_library, ONLY : g05kff, g05kjf, g05saf, nag_wp
       USE omp_lib, ONLY     : omp_get_thread_num, omp_get_num_threads
!      .. Implicit None Statement ..
       IMPLICIT NONE
!      .. Parameters ..
       INTEGER, PARAMETER              :: lseed = 1, nin = 5, nout = 6
       INTEGER, PARAMETER              :: wp = kind(0.0E0_nag_wp)
!      .. Local Scalars ..
       REAL (KIND=nag_wp)              :: sumrandom
       INTEGER                         :: genid, i, ifail, iskip, lstate,      &
                                          npaths, numberofthreads, subid, t1,  &
                                          threadno, tnpaths, tskip
!      .. Local Arrays ..
       REAL (KIND=nag_wp), ALLOCATABLE :: x(:)
       INTEGER                         :: seed(lseed)
       INTEGER, ALLOCATABLE            :: state(:)
!      .. Intrinsic Functions ..
       INTRINSIC                          kind, mod, sum
!      .. Executable Statements ..
       CONTINUE

!      skip the header in the data file
       READ (nin,*)

!      use the Wichmann Hill II generator:
!      at Mark 22 of the Fortran library skip ahead is only available for 4 of the
!      available 5 generators, the exception being the Mersenne Twister. Skip 
!      ahead for the Mersenne Twister will appear in the next Mark
       genid = 4
       subid = 0

!      read in the problem size
!      where NPATHS = number of paths (values), ISKIP = initial skip and SEED(1)
!      is a seed value
       READ (nin,*) npaths, iskip, seed(1)

!      in this example we are going to be setting IFAIL = 0 prior to calling any
!      NAG routine this indicates a hard failure if any errors or warnings are 
!      returned by the NAG routine which means the calling program will terminate.
!      In more complicated applications a value of 1 may be more appropriate (see
!      the essential introduction for more details). In such cases the return 
!      value of the IFAIL parameter should be checked

!      get the length of the state array for the psuedo-generator
       lstate = -1
       ifail = 0
       CALL g05kff(genid,subid,seed,lseed,state,lstate,ifail)

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP     REDUCTION(+:SUMRANDOM), &
!$OMP     PRIVATE(THREADNO,TSKIP,TNPATHS,X,IFAIL,I,T1,NUMBEROFTHREADS,STATE), &
!$OMP     SHARED(NPATHS,ISKIP,GENID,SUBID,LSTATE,SEED)

!      thread number
       threadno = omp_get_thread_num()

!      number of threads
       numberofthreads = omp_get_num_threads()

!      calculate the skip and the number of values to generate for this thread 
       IF (numberofthreads==1) THEN
          tskip = 0
          tnpaths = npaths
       ELSE
          t1 = mod(npaths,numberofthreads)
          IF (threadno<t1) THEN
             t1 = threadno
          END IF
          tskip = threadno*(npaths/numberofthreads) + t1
          tnpaths = (npaths+numberofthreads-threadno-1)/numberofthreads
       END IF

!      adjust the skip for this thread by the initial skip
       tskip = tskip + iskip

!      allocate memory to hold the generated values (X) and state of the pseudo
!      generator (STATE)
       ALLOCATE (x(tnpaths),state(lstate))

!      initialise the pseudo generator, all threads need to be initialised using
!      the same seed
       ifail = 0
       CALL g05kff(genid,subid,seed,lseed,state,lstate,ifail)

!      skip ahead the sequence tskip places on each thread
       ifail = 0
       CALL g05kjf(tskip,state,ifail)

!      generate tnpaths values from a uniform distribution 
       ifail = 0
       CALL g05saf(tnpaths,state,x,ifail)

!      calculate the sum for this particular thread */    
       sumrandom = sum(x(1:tnpaths))

       WRITE (nout,'(A,I4,A,E24.14)') 'Thread = ', threadno + 1, ' Sum = ', &
          sumrandom

!      free memory
       DEALLOCATE (x,STAT=i)
       DEALLOCATE (state,STAT=i)
!$OMP END PARALLEL

!      display the results
       WRITE (nout,'(A,E24.14)') 'Total Sum = ', sumrandom

    END PROGRAM fpseudo_openmp
