    PROGRAM fscrambled_openmp
!      NAG Copyright 2011.
!      an example of calling the NAG Sobol G05YMF with scrambled sequences in an
!      OpenMP multi-threaded environment

!      .. Use Statements ..
       USE nag_library, ONLY : g05kff, g05ymf, g05ynf, nag_wp
       USE omp_lib, ONLY     : omp_get_thread_num, omp_get_num_threads
!      .. Implicit None Statement ..
       IMPLICIT NONE
!      .. Parameters ..
       INTEGER, PARAMETER              :: lseed = 1, nin = 5, nout = 6
       INTEGER, PARAMETER              :: wp = kind(0.0E0_nag_wp)
!      .. Local Scalars ..
       REAL (KIND=nag_wp)              :: sumrandom
       INTEGER                         :: genid, i, idim, ifail, iskip,        &
                                          ldquas, liref, lstate, npaths,       &
                                          nsdigi, numberofthreads, pgenid,     &
                                          psubid, rcord, sdquas, stype, t1,    &
                                          threadno, tnpaths, tskip
!      .. Local Arrays ..
       REAL (KIND=nag_wp), ALLOCATABLE :: quas(:,:)
       INTEGER, ALLOCATABLE            :: iref(:), state(:)
       INTEGER                         :: seed(lseed)
!      .. Intrinsic Functions ..
       INTRINSIC                          kind, mod, sum
!      .. Executable Statements ..
       CONTINUE

!      skip the header in the data file
       READ (nin,*)

!      we are currently just summing the values for the first dimension, so makes
!      sense to work in with RCORD = 2, where all values from a given dimension 
!      are contiguous
       rcord = 2

!      read in the problem size
!      where NPATHS = number of paths (values), IDIM = number of dimensions, 
!      ISKIP = initial skip and SEED(1) is a seed value
       READ (nin,*) npaths, idim, iskip, seed(1)

!      want the Sobol generator and Owen-like scrambling 
       genid = 1
       stype = 1

!      scramble the default number of digits
       nsdigi = 0

!      use the Mersenne Twister psuedo-generator
       pgenid = 3
       psubid = 0

!      in this example we are going to be setting IFAIL = 0 prior to calling any 
!      NAG routine this indicates a hard failure if any errors or warnings are 
!      returned by the NAG routine which means the calling program will terminate.
!      In more complicated applications a value of 1 may be more appropriate (see
!      the essential introduction for more details). In such cases the return 
!      value of the IFAIL parameter should be checked

!      get the length of the state array for the psuedo-generator
       lstate = -1
       ifail = 0
       CALL g05kff(pgenid,psubid,seed,lseed,state,lstate,ifail)

!      calculate the size of the reference vector
       liref = 32*idim + 7

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP     REDUCTION(+:SUMRANDOM), &
!$OMP     PRIVATE(THREADNO,TSKIP,TNPATHS,QUAS,IREF,IFAIL,I,T1,LDQUAS,SDQUAS,&
!$OMP             NUMBEROFTHREADS,STATE), &
!$OMP     SHARED(NPATHS,IDIM,LIREF,ISKIP,GENID,RCORD,PGENID,PSUBID,LSTATE,SEED,&
!$OMP            NSDIGI,STYPE)

!      thread number
       threadno = omp_get_thread_num()

!      number of threads
       numberofthreads = omp_get_num_threads()

!      calculate the skip and the number of values to generate for this thread 
       IF (numberofthreads==1) THEN
          tskip = 0
          tnpaths = npaths
       ELSE
          t1 = mod(npaths,numberofthreads)
          IF (threadno<t1) THEN
             t1 = threadno
          END IF
          tskip = threadno*(npaths/numberofthreads) + t1
          tnpaths = (npaths+numberofthreads-threadno-1)/numberofthreads
       END IF

!      adjust the skip for this thread by the initial skip
       tskip = tskip + iskip

!      calculate the lead and trailing dimension of QUAS
       IF (rcord==1) THEN
          ldquas = idim
          sdquas = tnpaths
       ELSE
          ldquas = tnpaths
          sdquas = idim
       END IF

!      allocate memory to hold the generated values (QUAS), state of the 
!      quasi-random generator (IREF) and state of the pseudo generator (STATE)
       ALLOCATE (quas(ldquas,sdquas),state(lstate),iref(liref))

!      initialise the pseudo generator, all threads need to be initialised using
!      the same seed as we want to use the same scrambling on each thread, each 
!      thread needs its own copy of state as this is input / output to NAG routine
       ifail = 0
       CALL g05kff(pgenid,psubid,seed,lseed,state,lstate,ifail)

!      initialise the quasi generator */
       ifail = 0
       CALL g05ynf(genid,stype,idim,iref,liref,tskip,nsdigi,state,ifail)

!      generate the required values
       ifail = 0
       CALL g05ymf(tnpaths,rcord,quas,ldquas,iref,ifail)

!      calculate the sum for this particular thread. We are only interested in 
!      the first dimension which are stored in the first TNPATHS of the QUAS array
       sumrandom = sum(quas(1:tnpaths,1))

       WRITE (nout,'(A,I4,A,E24.14)') 'Thread = ', threadno + 1, ' Sum = ', &
          sumrandom

!      free memory
       DEALLOCATE (quas,STAT=i)
       DEALLOCATE (iref,STAT=i)
       DEALLOCATE (state,STAT=i)
!$OMP END PARALLEL

!      display the results
       WRITE (nout,'(A,E24.14)') 'Total Sum = ', sumrandom

    END PROGRAM fscrambled_openmp
